/*
 * #{copyright}#
 */
package com.hand.hap.account.controllers;

import com.hand.hap.account.constants.UserConstants;
import com.hand.hap.account.dto.RoleExt;
import com.hand.hap.account.dto.User;
import com.hand.hap.account.dto.UserRole;
import com.hand.hap.account.exception.UserException;
import com.hand.hap.account.mapper.UserMapper;
import com.hand.hap.account.service.IRoleService;
import com.hand.hap.account.service.IUserInfoService;
import com.hand.hap.account.service.IUserService;
import com.hand.hap.core.IRequest;
import com.hand.hap.core.exception.BaseException;
import com.hand.hap.security.PasswordManager;
import com.hand.hap.security.service.impl.DefaultUserSecurityStrategy;
import com.hand.hap.system.controllers.BaseController;
import com.hand.hap.system.dto.ResponseData;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Controller;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.servlet.ModelAndView;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

/**
 * UserController.
 *
 * @author njq.niu@hand-china.com
 */
@Controller
@RequestMapping(value = {"/sys/user", "/api/sys/user"})
public class UserController extends BaseController {

    @Autowired
    private IUserService userService;
    @Autowired
    @Qualifier("roleServiceImpl")
    private IRoleService roleService;

    @Autowired
    private IUserInfoService userInfoService;

    @Autowired
    private PasswordManager passwordManager;


    /**
     * 查询用户数据.
     *
     * @param user     用户
     * @param page     起始页
     * @param pagesize 分页大小
     * @return ResponseData
     */
    @RequestMapping(value = "/query", method = RequestMethod.POST)
    @ResponseBody
    public ResponseData selectUsers(HttpServletRequest request, @ModelAttribute User user, @RequestParam(defaultValue = DEFAULT_PAGE) int page, @RequestParam(defaultValue = DEFAULT_PAGE_SIZE) int pagesize) {
        IRequest iRequest = createRequestContext(request);
        List<User> usersList = userService.selectUsers(iRequest, user, page, pagesize);
        Date day=new Date();
        for(User u :usersList){
            //当前时间在到期日之后更新状态
            if(u.getEndActiveDate()!=null && day.after(u.getEndActiveDate())){
                u.setStatus("EXPR");
                userService.updateStatus(u.getUserId(),u.getStatus());
            };
            //当前时间在到期日之前更新状态
            if(u.getEndActiveDate()!=null && day.before(u.getEndActiveDate())){
                u.setStatus("ACTV");
                userService.updateStatus(u.getUserId(),u.getStatus());
            };
        }
        return new ResponseData(usersList);
    }

    /**
     * 保存更新账户数据.
     *
     * @param users   用户
     * @param result  BindingResult
     * @param request 请求上下文
     * @return ResponseData ResponseData
     * @throws BaseException BaseException
     */
    @RequestMapping(value = "/submit", method = RequestMethod.POST)
    @ResponseBody
    public ResponseData submitUsers(@RequestBody List<User> users, BindingResult result, HttpServletRequest request) throws BaseException {
        getValidator().validate(users, result);
        if (result.hasErrors()) {
            ResponseData rd = new ResponseData(false);
            rd.setMessage(getErrorMessage(result, request));
            return rd;
        }
        IRequest requestCtx = createRequestContext(request);
        userService.batchUpdate(requestCtx, users);
        return new ResponseData(users);

    }


    /**
     * 更新个人信息.
     *
     * @param request 请求上下文
     * @param user 用户
     * @return ResponseData ResponseData
     * @throws BaseException BaseException
     * /sys/um/updateUserInfo
     */
    @RequestMapping(value = "/update", method = RequestMethod.POST)
    public ResponseData updateUserInfo(HttpServletRequest request, @RequestBody User user) throws BaseException {
        IRequest iRequest = createRequestContext(request);

        userService.validateEmail(user.getEmail());
        userService.validatePhone(user.getPhone());
        user.setUserId(iRequest.getUserId());
        userInfoService.update(iRequest, user);
        return new ResponseData(Arrays.asList(user));
    }

    /**
     * 删除账户.
     *
     * @param users 用户列表
     * @return ResponseData ResponseData
     * @throws BaseException BaseException
     */
    @RequestMapping(value = "/remove", method = RequestMethod.POST)
    public ResponseData remove(@RequestBody List<User> users) throws BaseException {
        userService.batchDelete(users);
        return new ResponseData(users);
    }

    @RequestMapping(value = "/{userId}/roles", method = RequestMethod.POST)
    @ResponseBody
    public ResponseData selectUserAndRoles(HttpServletRequest request, @PathVariable Long userId) {
        IRequest iRequest = createRequestContext(request);
        ResponseData rd = new ResponseData();
        User user = new User();
        user.setUserId(userId);
        rd.setRows(roleService.selectRolesByUser(iRequest, user));
        return rd;
    }

    @RequestMapping(value = "/{userId}/roles/query", method = RequestMethod.POST)
    @ResponseBody
    public ResponseData queryUserAndRoles(HttpServletRequest request, @PathVariable Long userId) {
        IRequest iRequest = createRequestContext(request);
        ResponseData rd = new ResponseData();
        User user = new User();
        user.setUserId(userId);
        rd.setRows(roleService.queryRolesByUser(iRequest, user));
        return rd;
    }

    /**
     * 首次登陆以及密码过期重置密码.
     *
     * @param request
     * @param password
     * @return ResponseData
     * @throws UserException "/sys/um/resetPasswordAdmin"
     */
    @RequestMapping(value = "/password/reset", method = RequestMethod.POST)
    public ResponseData updatePassword(HttpServletRequest request, String password, String passwordAgain, Long userId) throws UserException {
        IRequest iRequest = createRequestContext(request);
        userService.validatePassword(iRequest, password, passwordAgain);
        userService.updatePassword(userId, password);
        userService.updateNotFirstLogin(userId, User.FIRST_LOGIN_STATUS);
        return new ResponseData(true);
    }


    /**
     * 密码更新.
     *
     * @param request
     * @param oldPwd
     * @param newPwd
     * @param newPwdAgain
     * @return
     * @throws UserException
     */
    @RequestMapping(value = "/password/update", method = RequestMethod.POST)
    public ResponseData updatePassword(HttpServletRequest request, String oldPwd, String newPwd, String newPwdAgain) throws UserException {
        IRequest iRequest = createRequestContext(request);
        Long accountId = iRequest.getUserId();
        HttpSession session = request.getSession(false);
        //TODO: expire??
        Object expire = session.getAttribute(DefaultUserSecurityStrategy.PASSWORD_UPDATE_REASON);
        iRequest.setAttribute(DefaultUserSecurityStrategy.PASSWORD_UPDATE_REASON, expire);
        userService.validatePassword(iRequest, newPwd, newPwdAgain);

        if (StringUtils.isEmpty(oldPwd)) {
            throw new UserException(UserException.PASSWORD_NOT_EMPTY, null);
        }
        User tmp = new User();
        tmp.setUserId(accountId);
        User userInDB = userService.selectByPrimaryKey(iRequest, tmp);
        String pwd = userInDB.getPasswordEncrypted();
        if (!passwordManager.matches(oldPwd, pwd)) {
            throw new UserException(UserException.USER_PASSWORD_WRONG, null);
        }

        if (passwordManager.matches(newPwd, pwd)) {
            throw new UserException(UserException.USER_PASSWORD_SAME, null);
        }

        userService.updatePassword(accountId, newPwd);
        // 如果用户密码过期 修改成功后 清除session
        if (expire != null) {
            session.removeAttribute(DefaultUserSecurityStrategy.PASSWORD_UPDATE_REASON);
            // 更新是否第一次登录的状态
            userService.updateNotFirstLogin(accountId, User.NOT_FIRST_LOGIN_STATUS);
        }
        return new ResponseData(true);
    }

    @RequestMapping(value = "/queryFunction", method = RequestMethod.POST)
    @ResponseBody
    public ResponseData queryFunction(HttpServletRequest request, @RequestParam(required = false) Long userId) {
        IRequest requestContext = createRequestContext(request);
        return new ResponseData(userService.queryFunction(requestContext, userId));
    }
}
