package com.hand.hap.excel.event;

import com.hand.hap.excel.dto.ExportConfig;
import com.hand.hap.excel.utils.ExcelUtils;
import com.hand.hap.excel.utils.SQLHelper;
import com.hand.hap.message.IMessageConsumer;
import com.hand.hap.message.IMessagePublisher;
import com.hand.hap.message.QueueMonitor;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.xssf.streaming.SXSSFSheet;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.HashOperations;
import org.springframework.data.redis.core.ListOperations;
import org.springframework.data.redis.core.SetOperations;

import javax.annotation.Resource;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicInteger;

@QueueMonitor(queue = "prepareExport")
public class PrepareExcelListener implements IMessageConsumer<String> {
    private int MAX_ROWS = 10000000;

    @Autowired
    @Qualifier("sqlSessionFactory")
    SqlSessionFactory sqlSessionFactory;

    @Resource(name = "v2redisTemplate")
    private ListOperations<String, String> listOperations;

    @Resource(name = "v2redisTemplate")
    private HashOperations<String, String, Object> hashOperations;

    @Resource(name = "v2redisTemplate")
    private SetOperations<String, String> setOperations;

    @Value("${export.filePath?:'c:\\'}")
    private String exportFilePath;

    @Autowired
    private IMessagePublisher messagePublisher;

    private String exportKey = "exporting";
    private String prepareKey = "prepareExport";
    private String readyExport = "readyExport";
    private Logger logger = LoggerFactory.getLogger(PrepareExcelListener.class);

    @Override
    public void onMessage(String message, String queue) {
        String userKey = message;
        String key = "";
        long inQueueTime = System.currentTimeMillis();
        try {
            ExportConfig config = (ExportConfig) hashOperations.get(userKey, "config");
            String sqlId = hashOperations.get(userKey, "sqlId").toString();
            String userId = hashOperations.get(userKey, "userId").toString();
            Object rowObject = hashOperations.get(userKey, "rowMaxNumber");
            Object dateObject = hashOperations.get(userKey, "startDate");
            String startDate = "";
            if (dateObject != null) {
                long date = Long.valueOf(dateObject.toString());
                startDate += ("---" + date);
            }
            key = userKey + "---" + inQueueTime + startDate;
            setOperations.add(exportKey, key);

            int rowMaxNumber = MAX_ROWS;
            if (rowObject != null) {
                rowMaxNumber = Integer.valueOf(rowObject.toString());
            }
            String savePath = exportExcel(sqlId, config, rowMaxNumber);
            setOperations.remove(exportKey, key);

            // 将filepath存到ready队列中
            String readyValue = userId
                    + "---" + config.getFileName()
                    + "---" + savePath
                    + "---" + inQueueTime
                    + startDate
                    + "---" + System.currentTimeMillis();

            hashOperations.delete(userKey, "config", "sqlId", "userId", "rowMaxNumber");
            listOperations.rightPush(readyExport, readyValue);

        } catch (Exception e) {
            // 出现了异常，需要把信息重新填回队列中
            setOperations.remove(exportKey, key);
            messagePublisher.message(prepareKey, userKey);
            logger.error("something error! \n userKEY：" + userKey);
            logger.error(e.getMessage());
            e.printStackTrace();
        }
    }

    private String exportExcel(String sqlId, ExportConfig gridInfo, int rowMaxNumber)
            throws IOException {
        String filePath = null;
        SXSSFWorkbook wb = new SXSSFWorkbook(1000);
        CellStyle dateFormat = wb.createCellStyle();
        dateFormat.setDataFormat(wb.createDataFormat().getFormat("yyyy-MM-DD HH:mm"));
        // row计数器
        final AtomicInteger count = new AtomicInteger(1);
        // sheet页row索引
        final AtomicInteger rowIndex = new AtomicInteger(1);

        ExcelUtils.initColumnType(gridInfo.getColumnsInfo(), gridInfo.getParam());

        final SXSSFSheet[] sheet = {wb.createSheet()};

        ExcelUtils.createHeaderRow(gridInfo.getColumnsInfo(), wb, sheet[0]);

        try (SqlSession sqlSession = sqlSessionFactory.openSession()) {
            sqlSession.select(sqlId, gridInfo.getParam(), (resultContext -> {
                Object object = resultContext.getResultObject();
                sheet[0] = ExcelUtils.createSheet(wb, sheet[0], object, count, rowIndex, rowMaxNumber, gridInfo, dateFormat);
            }));
            String namespaceSql = SQLHelper.getNamespaceSql(sqlSession, sqlId, gridInfo.getParam());
            System.out.println(namespaceSql);
            // 保存到导出路径下
            if (exportFilePath != null && !exportFilePath.isEmpty()) {
                // 先保存文件到指定路径下
                String uuid = UUID.randomUUID().toString();
                String fileName = "file" + uuid + ".xlsx";
                String savePath = exportFilePath + fileName;
                FileOutputStream fileOutputStream = new FileOutputStream(savePath);
                wb.write(fileOutputStream);
                filePath = savePath;
            }
        } finally {
            wb.close();
            wb.dispose();
        }
        return filePath;
    }
}
