package com.hand.hap.excel.service.impl;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.hand.hap.account.dto.Role;
import com.hand.hap.account.mapper.RoleMapper;
import com.hand.hap.core.IRequest;
import com.hand.hap.excel.dto.ExcelExportDTO;
import com.hand.hap.excel.dto.ExportConfig;
import com.hand.hap.excel.mapper.ExcelExportMapper;
import com.hand.hap.excel.service.IExcelExportService;
import com.hand.hap.excel.utils.SQLHelper;
import com.hand.hap.mybatis.entity.Example;
import com.hand.hap.system.dto.ResponseData;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.http.client.methods.HttpPost;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.net.URLEncoder;
import java.util.Date;
import java.util.List;

/**
 * Created by yang.yang07@hand-china.com on 2017/7/30.
 */
@Service
public class ExcelExportServiceImpl implements IExcelExportService {
    private static final String ENC = "UTF-8";

    private Logger logger = LoggerFactory.getLogger(getClass());

    @Autowired
    @Qualifier("sqlSessionFactory")
    SqlSessionFactory sqlSessionFactory;

    @Value("${export.offerUrl}")
    private String downloadUrl;

    @Autowired
    private ExcelExportMapper exportMapper;

    @Autowired
    private RoleMapper roleMapper;

    @Autowired
    ObjectMapper objectMapper;

    @Override
    public void saveExportInfo(String sqlId, IRequest iRequest, ExportConfig config, int rowMaxNumber) {
        ExcelExportDTO dto = new ExcelExportDTO();
        dto.setSqlId(sqlId);
        dto.setStatus("new");
        dto.setFileName(config.getFileName());
        dto.setInDate(new Date(System.currentTimeMillis()));
        dto.setEnable("Y");
        dto.setUserId(iRequest.getUserId());
        SqlSession sqlSession = sqlSessionFactory.openSession();
        String namespaceSql = SQLHelper.getNamespaceSql(sqlSession, sqlId, config.getParam());
        logger.info("=============sql id=======" + namespaceSql.length());
        logger.info(namespaceSql);

        dto.setSqlId(namespaceSql);
        try {
            dto.setExportJsonConfig(objectMapper.writeValueAsString(config));
        } catch (JsonProcessingException e) {
            logger.error(e.getMessage(), e);
        }
        logger.info("=============config =======\n" + dto.getExportJsonConfig());
        dto.setMaxRowNumber(rowMaxNumber);
        exportMapper.insertSelective(dto);
    }

    @Override
    public ResponseData queryExportInfo(String status, IRequest iRequest) {
        Example example = new Example(ExcelExportDTO.class);
        if (status != null) {
            example.createCriteria().andEqualTo("status", status);
        }
        example.createCriteria().andEqualTo("enable", "Y");
        example.setOrderByClause("status");
        Long userId = iRequest.getUserId();
        Long roleId = iRequest.getRoleId();
        Role role = new Role();
        role.setRoleId(roleId);
        role = roleMapper.selectByPrimaryKey(role);
        try {
            if (!"ADMIN".equals(role.getRoleCode())) {
                example.createCriteria().andEqualTo("userId", iRequest.getUserId());
            }
            List<ExcelExportDTO> dtos = exportMapper.selectByExample(example);
            return new ResponseData(dtos);
        } catch (Exception e) {
            logger.error("Excel文件导出信息查询失败", e);
        }
        return new ResponseData(false);
    }

    @Override
    public void downloadExcel(String filePath, String fileName, HttpServletRequest request, HttpServletResponse response) {

        logger.info("==========prepare download=====================");
        logger.info("file path: " + filePath);
        logger.info("file name: " + fileName);
        logger.info("download url: " + downloadUrl);
        HttpClient httpClient = new HttpClient();
        ServletOutputStream outputStream = null;
        try {
            outputStream = response.getOutputStream();
            PostMethod downloadPost = new PostMethod(downloadUrl);
            HttpPost httpPost = new HttpPost();

            // 准备参数
            downloadPost.setParameter("filePath", URLEncoder.encode(filePath, ENC));
            downloadPost.setParameter("fileName", URLEncoder.encode(fileName, ENC));
            logger.info("============parameter prepared===============");
            httpClient.executeMethod(downloadPost);

            int downloadStatus = downloadPost.getStatusCode();
            logger.info("============status code :" + downloadStatus);
            if (downloadStatus == HttpStatus.SC_NOT_FOUND) {
                throw new IOException("file not found,file path:" + filePath);
            } else if (downloadStatus == HttpStatus.SC_OK) {
                // 文件获取成功，开始处理文件数据
                try (InputStream in = downloadPost.getResponseBodyAsStream()) {
                    long contentLength = downloadPost.getResponseContentLength();
                    response.addHeader("Content-Disposition", "attachment;filename=\"" + URLEncoder.encode(fileName, ENC) + "\"");
                    response.setContentType("application/vnd.ms-excel" + ";charset=" + ENC);
                    response.setHeader("Accept-Ranges", "bytes");
                    response.setContentLength((int) contentLength);

                    byte flush[] = new byte[1024];
                    int len = 0;
                    while (0 <= (len = in.read(flush))) {
                        outputStream.write(flush, 0, len);
                        logger.info("=====write bytes: " + len);
                    }
                    logger.info("==========excel downloaded!==========");
                }
            }
        } catch (IOException e) {
            logger.error("download failed");
            logger.error(e.getMessage(), e);
        } finally {
            if (outputStream != null) {
                try {
                    outputStream.close();
                } catch (IOException e) {
                    logger.error(e.getMessage(), e);
                }
            }
        }
    }

    @Override
    public void cancelExport(Long id) {
        if (id == null) {
            return;
        }
        ExcelExportDTO dto = exportMapper.selectByPrimaryKey(id);
        if (dto != null) {
            exportMapper.deleteByPrimaryKey(dto);
        }
    }

    @Override
    public void exportExcel(String sqlId, ExportConfig exportConfig, HttpServletRequest httpServletRequest,
                            HttpServletResponse httpServletResponse, IRequest iRequest, int rowMaxNumber) throws IOException {
        saveExportInfo(sqlId, iRequest, exportConfig, rowMaxNumber);
    }

    @Override
    public void exportExcel(String sqlId, ExportConfig exportConfig, HttpServletRequest httpServletRequest,
                            HttpServletResponse httpServletResponse, IRequest iRequest) throws IOException {
        exportExcel(sqlId, exportConfig, httpServletRequest, httpServletResponse, iRequest, 1000000);
    }

}
