package com.hand.hap.excel.service.impl;

import com.hand.hap.core.IRequest;
import com.hand.hap.excel.dto.ColumnInfo;
import com.hand.hap.excel.dto.ExportConfig;
import com.hand.hap.excel.service.IExportService;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.collections.CollectionUtils;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.HorizontalAlignment;
import org.apache.poi.ss.usermodel.VerticalAlignment;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.xssf.streaming.SXSSFCell;
import org.apache.poi.xssf.streaming.SXSSFRow;
import org.apache.poi.xssf.streaming.SXSSFSheet;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.util.ReflectionUtils;
import org.springframework.util.StringUtils;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.util.Date;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Created by jialong.zuo@hand-china.com on 2016/11/30.
 */
@Service
public class ExportServiceImpl implements IExportService {

    private static final String ENC = "UTF-8";

    private Logger logger = LoggerFactory.getLogger(getClass());

    @Autowired
    @Qualifier("sqlSessionFactory")
    SqlSessionFactory sqlSessionFactory;


    private void exportExcel(String sqlId, ExportConfig gridInfo, IRequest iRequest, OutputStream outputStream, int rowMaxNumber)
            throws IOException {
        SXSSFWorkbook wb = new SXSSFWorkbook(50);
        CellStyle dateFormat = wb.createCellStyle();
        dateFormat.setDataFormat(wb.createDataFormat().getFormat("yyyy-MM-DD HH:mm"));
        // row计数器
        final AtomicInteger count = new AtomicInteger(1);
        // sheet页row索引
        final AtomicInteger rowIndex = new AtomicInteger(1);

        initColumnType(gridInfo.getColumnsInfo(), gridInfo.getParam());

        final SXSSFSheet[] sheet = {wb.createSheet()};

        String title = gridInfo.getTitle();
        if(StringUtils.isEmpty(title)) {
            createHeaderRow(gridInfo.getColumnsInfo(), wb, sheet[0]);
        }else{
            rowIndex.getAndIncrement();
            createTitleRow(title,wb,sheet[0],gridInfo.getColumnsInfo());
            createHeaderRow(gridInfo.getColumnsInfo(), wb, sheet[0],1);
        }
        try (SqlSession sqlSession = sqlSessionFactory.openSession()) {
            sqlSession.select(sqlId, gridInfo.getParam(), (resultContext -> {
                Object object = resultContext.getResultObject();
                sheet[0] = createSheet(wb, sheet[0], object, count, rowIndex, rowMaxNumber, gridInfo, dateFormat);
            }));
            wb.write(outputStream);
        } finally {
            wb.close();
            wb.dispose();
        }
    }

    private void createTitleRow(String title, SXSSFWorkbook wb, SXSSFSheet rows, List columnsInfo) {
        if(!StringUtils.isEmpty(title) && CollectionUtils.isNotEmpty(columnsInfo)){
            SXSSFRow row = rows.createRow(0);
            SXSSFCell cell = row.createCell(0);
            cell.setCellValue(title);
            CellStyle cellStyle = wb.createCellStyle();
            cellStyle.setVerticalAlignment(VerticalAlignment.CENTER);
            cellStyle.setAlignment(HorizontalAlignment.CENTER);
            cell.setCellStyle(cellStyle);
            CellRangeAddress cellRangeAddress = new CellRangeAddress(0, 0, 0, columnsInfo.size() - 1);
            rows.addMergedRegion(cellRangeAddress);
        }
    }

    private SXSSFSheet createSheet(SXSSFWorkbook wb, SXSSFSheet sheet, Object object, AtomicInteger count, AtomicInteger rowIndex, int rowMaxNumber, ExportConfig gridInfo, CellStyle dateFormat) {
        if (count.get() % rowMaxNumber == 0) {
            sheet = wb.createSheet();
            createHeaderRow(gridInfo.getColumnsInfo(), wb, sheet);
            rowIndex.set(0);
        }
        count.getAndIncrement();
        SXSSFRow row = sheet.createRow(rowIndex.getAndIncrement());
        createRow(gridInfo.getColumnsInfo(), object, row, dateFormat);
        return sheet;
    }


    private void createRow(List<ColumnInfo> columnInfos, Object object, SXSSFRow row, CellStyle dateFormat) {
        for (int ii = 0; ii < columnInfos.size(); ii++) {
            Object fieldObject = null;
            try {
                fieldObject = PropertyUtils.getProperty(object, columnInfos.get(ii).getName());
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
            String type = columnInfos.get(ii).getType();

            SXSSFCell cell = row.createCell(ii);

            if (null == fieldObject) {
                cell.setCellType(XSSFCell.CELL_TYPE_STRING);
                cell.setCellValue((String) null);
            } else {
                switch (type.toUpperCase()) {
                    case "NUMBER":
                    case "FLOAT":
                        cell.setCellType(XSSFCell.CELL_TYPE_NUMERIC);
                        cell.setCellValue((Float) fieldObject);
                        break;
                    case "DOUBLE":
                        cell.setCellType(XSSFCell.CELL_TYPE_NUMERIC);
                        cell.setCellValue((Double) fieldObject);
                        break;
                    case "INT":
                    case "INTEGER":
                    case "LONG":
                        cell.setCellType(XSSFCell.CELL_TYPE_NUMERIC);
                        cell.setCellValue((Long) fieldObject);
                        break;
                    case "DATE":
                        cell.setCellStyle(dateFormat);
                        cell.setCellValue((Date) fieldObject);
                        break;
                    case "BOOLEAN":
                        cell.setCellType(XSSFCell.CELL_TYPE_BOOLEAN);
                        cell.setCellValue((Boolean) fieldObject);
                        break;
                    default:
                        cell.setCellType(XSSFCell.CELL_TYPE_STRING);
                        cell.setCellValue((String) fieldObject);
                        break;
                }
            }
        }
    }

    private void createHeaderRow(List<ColumnInfo> columnInfos, SXSSFWorkbook wb, SXSSFSheet sheet) {
        createHeaderRow(columnInfos,wb,sheet,0);
    }

    private void createHeaderRow(List<ColumnInfo> columnInfos, SXSSFWorkbook wb, SXSSFSheet sheet, int rowIndex) {
        SXSSFRow firstRow = sheet.createRow(rowIndex);
        // 设置列字体align
        CellStyle cellStyle = wb.createCellStyle();
        cellStyle.setAlignment(CellStyle.ALIGN_CENTER);
        for (int i = 0; i < columnInfos.size(); i++) {
            SXSSFCell firstCell = firstRow.createCell(i);
            firstCell.setCellValue(columnInfos.get(i).getTitle());
            // 设置列宽度
            sheet.setColumnWidth(i, columnInfos.get(i).getWidth() * 80);

            firstCell.setCellStyle(cellStyle);
        }
    }

    private void initColumnType(List<ColumnInfo> columnInfos, Object object) {
        for (ColumnInfo columnInfo : columnInfos) {
            columnInfo.setType(ReflectionUtils.findField(object.getClass(), columnInfo.getName()).getType().getSimpleName());
        }

    }

    @Override
    public void exportAndDownloadExcel(String sqlId, ExportConfig exportConfig, HttpServletRequest httpServletRequest,
                                       HttpServletResponse httpServletResponse, IRequest iRequest, int rowMaxNumber) throws IOException {
        httpServletResponse.addHeader("Content-Disposition",
                "attachment;filename=\"" + URLEncoder.encode(exportConfig.getFileName() + ".xlsx", ENC) + "\"");
        httpServletResponse.setContentType("application/vnd.ms-excel" + ";charset=" + ENC);
        httpServletResponse.setHeader("Accept-Ranges", "bytes");
        try (OutputStream outputStream = httpServletResponse.getOutputStream()) {
            exportExcel(sqlId, exportConfig, iRequest, outputStream, rowMaxNumber);
        }
    }

    @Override
    public void exportAndDownloadExcel(String sqlId, ExportConfig exportConfig, HttpServletRequest httpServletRequest,
                                       HttpServletResponse httpServletResponse, IRequest iRequest) throws IOException {
        exportAndDownloadExcel(sqlId, exportConfig, httpServletRequest, httpServletResponse, iRequest, 1000000);
    }


}
