package com.hand.hap.message.websocket;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.hand.hap.message.IMessageConsumer;
import com.hand.hap.message.TopicMonitor;
import com.hand.hap.security.CustomUserDetails;
import hls.core.websocket.components.SysNoticeQueueInit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.util.StringUtils;
import org.springframework.web.socket.TextMessage;
import org.springframework.web.socket.WebSocketSession;

import java.io.IOException;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * modify by fjm 添加userId为key的websocketSessionMap
 *
 * @author njq.niu@hand-china.com
 */
@TopicMonitor(channel = {WebSocketSessionManager.CHANNEL_WEB_SOCKET})
public class WebSocketSessionManager implements IMessageConsumer<CommandMessage> {

    public static final String CHANNEL_WEB_SOCKET = "webSocket";

    private final Logger logger = LoggerFactory.getLogger(WebSocketSessionManager.class);

    private final Map<String, List<WebSocketSession>> webSocketSessionMap = new ConcurrentHashMap<>();

    private final Map<Long, List<WebSocketSession>> webSocketSessionUserIdMap = new ConcurrentHashMap<>();

    @Autowired
    private ObjectMapper objectMapper;

    @Autowired
    private SysNoticeQueueInit sysNoticeQueueInit;


    public void addSession(WebSocketSession session) {
        Principal principal = session.getPrincipal();
        String userName = principal.getName();
        Long userId = unwrapUserDetail(principal);
        if (!StringUtils.isEmpty(userName)) {
            List<WebSocketSession> sessions = webSocketSessionMap.get(userName);
            if (sessions == null) {
                sessions = new ArrayList<>();
                webSocketSessionMap.put(userName, sessions);
            }
            sessions.add(session);
            if (logger.isDebugEnabled()) {
                logger.debug("Add webSocketSession {}, total size {}", session.getId(), sessions.size());
            }
        }

        if (userId != null) {
            //施加同步锁
            synchronized (WebSocketSessionManager.class) {
                List<WebSocketSession> sessions = webSocketSessionUserIdMap.get(userId);
                if (sessions == null || sessions.isEmpty()) {
                    if (sessions == null)
                        sessions = new ArrayList<>();
                    webSocketSessionUserIdMap.put(userId, sessions);
                    //将任务放入线程池中执行
                    ExecutorService executorService = sysNoticeQueueInit.getExecutorService();
                    Map<Long, SysNoticeQueueInit.MonitorTask> monitorTaskMap = sysNoticeQueueInit.getMonitorTaskMap();
                    SysNoticeQueueInit.MonitorTask monitorTask = monitorTaskMap.get(userId);
                    if (monitorTask == null) {//如果根据userId获取到的task为空,则说明该用户为新建的用户,所以此时需要创建一个新的任务,并添加到monitorTaskMap中
                        //对线程池进行适当的扩容
                        monitorTask = sysNoticeQueueInit.createMonitorTask(userId);
                        int maximumPoolSize = ((ThreadPoolExecutor) executorService).getMaximumPoolSize();
                        int corePoolSize = ((ThreadPoolExecutor) executorService).getCorePoolSize();
                        ((ThreadPoolExecutor) executorService).setMaximumPoolSize(maximumPoolSize + 1);
                        ((ThreadPoolExecutor) executorService).setCorePoolSize(corePoolSize + 1);

                    }
                    executorService.execute(monitorTask);
                }
                sessions.add(session);
            }
        }
    }


    public void removeSession(WebSocketSession session) {
        Principal principal = session.getPrincipal();
        String userName = principal.getName();
        Long userId = unwrapUserDetail(principal);
        if (!StringUtils.isEmpty(userName)) {
            List<WebSocketSession> sessions = webSocketSessionMap.get(userName);
            if (sessions != null) {
                sessions.remove(session);
                logger.debug("Remove webSocketSession {}, total size {}", session.getId(), sessions.size());
            }
        }
        if (userId != null) {
            List<WebSocketSession> sessions = webSocketSessionUserIdMap.get(userId);
            if (sessions != null) {
                sessions.remove(session);
            }
        }
        //如果当前用户没有使用websocket连接,那么从现场池中移除任务,停止对队列的监听
        if (webSocketSessionUserIdMap.get(userId) == null || webSocketSessionUserIdMap.get(userId).isEmpty()) {
            ExecutorService executorService = sysNoticeQueueInit.getExecutorService();
            Map<Long, SysNoticeQueueInit.MonitorTask> monitorTaskMap = sysNoticeQueueInit.getMonitorTaskMap();
            SysNoticeQueueInit.MonitorTask monitorTask = monitorTaskMap.get(userId);
            ((ThreadPoolExecutor) executorService).remove(monitorTask);
            monitorTask.stop();
        }
    }

    public List<WebSocketSession> getSession(String userName) {
        List<WebSocketSession> sessions = webSocketSessionMap.get(userName);
        return sessions != null ? sessions : Collections.EMPTY_LIST;
    }

    public List<WebSocketSession> getSession(Long userId) {
        List<WebSocketSession> sessions = webSocketSessionUserIdMap.get(userId);
        return sessions != null ? sessions : Collections.EMPTY_LIST;
    }

    public void sendCommandMessage(WebSocketSession session, CommandMessage commandMessage) {
        try {
            session.sendMessage(new TextMessage(objectMapper.writeValueAsString(commandMessage)));
        } catch (IOException e) {
            if (logger.isErrorEnabled()) {
                logger.error("send commandMessage error! commandMessage: {} ", commandMessage.toString());
            }
        }
    }


    @Override
    public void onMessage(CommandMessage commandMessage, String channel) {
        List<WebSocketSession> sessions = getSession(commandMessage.getUserName());
        if (sessions != null) {
            String sessionId = commandMessage.getSessionId();
            sessions.stream().filter(ws -> !sessionId.equalsIgnoreCase((String) ws.getAttributes().get("sessionId"))).forEach(webSocketSession -> {
                this.sendCommandMessage(webSocketSession, commandMessage);
            });
        }
    }

    /**
     * 解析userId
     *
     * @param principal
     * @return
     */
    private Long unwrapUserDetail(Principal principal) {
        if (principal instanceof UsernamePasswordAuthenticationToken) {
            Object principalUser = ((UsernamePasswordAuthenticationToken) principal).getPrincipal();
            if (principalUser instanceof CustomUserDetails) {
                Long userId = ((CustomUserDetails) principalUser).getUserId();
                return userId;
            }
        }
        return null;
    }
}
